"""Top-level package for OCTIS."""
from octis.dashboard import *
from octis.configuration import *
from octis.dataset import *
from octis.evaluation_metrics import *
from octis.models import *
from octis.optimization import *
from octis.preprocessing import *

__author__ = """Silvia Terragni"""
__email__ = 's.terragni4@campus.unimib.it'
__version__ = '1.6.0'
